# Sichel's generalized inverse Gaussian Poisson (GIGP)
# a = Inf, b = 0 is Poisson with mean a*b/2

# uses recurrence to evaluate density up to length-1
gigp <- function(length,a,b,g,log=F) {
  # a is alpha
  # b is theta
  # g is gamma
  if(a <= 0) stop("must have a > 0")
  if(b <= 0 || b >= 1) stop("must have 0 < b < 1")
  # base case
  x = 0:1
  p = g/2*log(1-b) -log(besselK(a*sqrt(1-b),g,exp=T))+a*sqrt(1-b)
  p = p +x*log(a*b/2) -lgamma(x+1) +log(besselK(a,x+g,exp=T))-a
  p = exp(p)
  # recurrence
  c2 = (a*b)^2/4
  for(i in 3:length) {
    x = i-1
    p[i] = b*(1+(g-1)/x)*p[i-1] + c2/x/(x-1)*p[i-2]
  }
  if(log) p = log(p)
  p
}

dgigp <- function(x,a,b,g=-0.5,log=F) {
  p = gigp(max(x)+1,a,b,g,log)
  p[x+1]
}

gigp.mean <- function(a,b,g=-0.5) {
  m = a*b/2/sqrt(1-b)
  if(g != -0.5) m = m * besselK(a*sqrt(1-b),1+g)/besselK(a*sqrt(1-b),g)
  m
}

gigp.mean.fit <- function(m,a,b,g=-0.5) {
  if(g == -0.5) {
    if(!missing(a)) {
      q = 2*m/a
      b = q*(sqrt(q^2+4)-q)/2
      b
    } else {
      a = m*2*sqrt(1-b)/b
      a
    }
  } else {
    gigp.mean.fcn <- function(b) (gigp.mean(a,b,g) - m)^2
    optimize(gigp.mean.fcn,c(0,1),tol=1e-8)$minimum
  }
}

if(F) {
  # test
  library(nupois)
  x <- (1:5000)-1
  a = .645
  #a = 10000
  g = -0.5
  #g = 0
  m = 6.872
  b = gigp.mean.fit(m,a=a,g=g)
  p = dgigp(x,a,b,g)
  print(p[1:5]*1083)
  plot(p[1:50],type="l")
  cat("sum(p)=",sum(p),"\n")
  p = p/sum(p)
  m = sum(p*x)
  m.true = gigp.mean(a,b,g)
  cat("true mean =",m.true,"est mean =",m,"\n")
  v <- sum(p*x*x) - m^2
  cat("true var =",NA,"est var =",v,"\n")
}
