# Morlat-Poisson(lambda,a)
# approx same as nu-poisson(lambda^a,a)
# a=1 is Poisson
# T is a*lambda

dmorlat <- function(x,lambda,a,log=F) {
  p = x
  i.z = which(x == 0)
  p[i.z] = 1 - pgamma(a*lambda,a)
  i.nz = which(x > 0)
  p[i.nz] = pgamma(a*lambda,a*x[i.nz]) - pgamma(a*lambda,a*(x[i.nz]+1))
  if(log) log(p) else p
}

morlat.mean <- function(lambda,a) {
  # similar to the approx for NuP mean
  lambda - (a-1)/2/a
}

# cpo randomly keeps 1/k of the arrivals in a Poisson process
# not same as keeping every kth arrival, as in morlat
# the results are similar to morlat, but integer constraint makes it more limited.

dcpo <- function(x,m,k,log=F) {
  a = m*k
  if(floor(k) != k) stop("k must be integer")
  p = rep(0,length(x))
  i = which(x == 0)
  for(j in seq(0,k-1)) {
    p[i] = p[i] + (1-j/k)*exp(-a)*a^j/gamma(j+1)
  }
  i = which(x > 0)
  for(j in seq(-k+1,k-1)) {
    p[i] = p[i] + (1-abs(j)/k)*exp(-a)*a^(k*x[i]+j)/gamma(k*x[i]+j+1)
  }
  if(log) log(p) else p
}

cpo.mean <- function(m,k) {
  m
}

dcnb <- function(x,a,b,k=1,log=F) {
  q = 1/(b+1)
  lq = log(q)
  l1q = log(1-q)
  p = rep(0,length(x))
  i = which(x == 0)
  for(j in seq(0,k-1)) {
    ch = lgamma(a+j)-lgamma(a)-lgamma(j+1)
    p[i] = p[i] + (1-j/k)*exp(ch + a*lq + j*l1q)
  }
  i = which(x > 0)
  for(j in seq(-k+1,k-1)) {
    kx = k*x[i]
    # choose() doesn't work
    ch = lgamma(a+kx+j)-lgamma(a)-lgamma(kx+j+1)
    p[i] = p[i] + (1-abs(j)/k)*exp(ch + a*lq + (kx+j)*l1q)
  }
  if(any(!is.finite(p))) {
    print(c(a,b,k))
    stop("dcnb returned nan")
  }
  if(log) log(p) else p
}

cnb.fit <- function(x) {
  dx = x/sum(x)
  dx[dx==0] = 1e-16
  i = seq(0,length(x)-1)
  cost.fcn <- function(theta,k) {
    a = theta[1]
    b = theta[2]
    if(a <= 0 || b <= 0) return(1e10)
    d = dcnb(i,a,b,k)+1e-16
    sum(x*log(dx/d))
  }
  # start with negbin
  theta = unlist(negbin.fit(i,w=x))
  best = Inf
  for(k in 1:20) {
    r = nlm(cost.fcn, theta[1:2], k=k)
    if(r$minimum < best) {
      best = r$minimum
      theta = c(r$estimate,k)
    } else break
  }
  list(a=theta[1],b=theta[2],k=theta[3])
}

if(F) {
  # test
  x = 0:100
  a = 0.1
  #a = 2
  lambda = 2.5
  p = dmorlat(x,lambda,a)
  #p = dcpo(x,lambda,a)
  plot(x,p,type="l")
  print(sum(p))
  m = sum(x*p)
  m.true = morlat.mean(lambda,a)
  cat("true mean =",m.true,"est mean =",m,"\n")

  if(T) {
    n = length(p)
    x1 = x[1:(n-1)]
    x2 = x[2:n]
    q = p[1:(n-1)]/p[2:n]
    plot(x1,q,type="l")
    plot(x1,log(q)/log(x2),type="l")
  }
}
