#include <math.h>

/* R extensions only return values by modifying arguments */

double logsumexp(double a, double b)
{
  double r;
  if(a > b) r = a + log(1 + exp(b-a));
  else r = b + log(1 + exp(a-b));
  return(r);
}

void jsd(int *n, double *lp1, double *lp2, double *result)
{
  double h1=0,h2=0,h12=0;
  double p1,p2;
  int i;
  for(i=0;i<*n;i++) {
    double lp12 = logsumexp(lp1[i],lp2[i])-M_LN2;
    double p12 = exp(lp12);
    if(p12 > 0) h12 += p12*lp12;
    p1 = exp(lp1[i]);
    if(p1 > 0) h1 += p1*lp1[i];
    p2 = exp(lp2[i]);
    if(p2 > 0) h2 += p2*lp2[i];
  }
  *result = 0.5*(h1 + h2) - h12;
}

#define NITER 10000

/* Computes the log of the normalizer for the COM-Poisson.
 * On exit, *lambda is the value, and *nu is the number of terms required
 * in the expansion.
 */
void compoisson_z_ln(double *lambda, double *nu)
{
  double tol = 1e-10;
  double lambda_ln;
  double z = log(1+*lambda), a, r;
  int i=2;
  if(*lambda <= 0) {
    *nu = 0;
    return;
  }
  if(*nu == 0) {
    /* special case */
    /* needed when lambda is close to 1 */
    *lambda = -log(1-*lambda);
    return;
  }
  a = lambda_ln = log(*lambda);
  for(;i<NITER;i++) {
    double e = lambda_ln - *nu*log(i);
    a = a+e;
    z = logsumexp(z,a);
    if(e < 0) {
      /* truncation error */
      r = exp(a-z)/(1-exp(e));
      /* truncation error in log domain */
      /* log(z+r)-log(z) = log(1 + r/z) =approx r/z */
      if(r < tol) break;
    }
  }
  if(i == NITER) {
    r = pow(*lambda,1/(*nu));
    if(r > 10) {
      /* use asymptotic approximation */
      z = (*nu)*r - (*nu-1)/2/(*nu)*lambda_ln - 
	(*nu-1)/2*log(2*M_PI) - 0.5*log(*nu);
      i = 0;
    }
  }
  *lambda = z;
  *nu = i;
}

/* Computes the log of the unnormalized pth moment (subtract Z to get the
 * moment).
 * On exit, *lambda is the value, and *nu is the number of terms required
 * in the expansion.
 */
void compoisson_zx_ln(double *lambda, double *nu, double *p)
{
  double tol = 1e-10;
  double lambda_ln = log(*lambda);
  double z = lambda_ln, a = lambda_ln, r;
  int i=2;
  if(*p == 0) { compoisson_z_ln(lambda,nu); return; }
  if(*lambda == 0) {
    *lambda = z;
    *nu = 0;
    return;
  }
  if(*nu == 0) {
    if(*p == 1) {
      *lambda = lambda_ln - 2*log(1-*lambda);
      return;
    } else if(*p == 2) {
      *lambda = lambda_ln + log(1+*lambda) - 3*log(1-*lambda);
      return;
    }
  }
  for(;i<NITER;i++) {
    double e = lambda_ln - *nu*log(i) + *p*log((double)i/(i-1));
    a = a+e;
    z = logsumexp(z,a);
    if(e < 0) {
      /* truncation error */
      r = exp(a-z)/(1-exp(e));
      /* truncation error in log domain */
      /* log(z+r)-log(z) = log(1 + r/z) =approx r/z */
      if(r < tol) break;
    }
  }
  if(i == NITER) {
    r = pow(*lambda,1/(*nu));
    if(r > 10) {
      /* use asymptotic approximation */
      z = (*nu)*r - (*nu-1)/2/(*nu)*lambda_ln - 
	(*nu-1)/2*log(2*M_PI) - 0.5*log(*nu);
      z = z + log(r - (*nu-1)/2/(*nu));
      i = 0;
    }
  }
  *lambda = z;
  *nu = i;
}

/*************************************************************************/

/* A less accurate way to compute Z. */
void compoisson_z(double *lambda, double *nu)
{
  double tol = 1e-8;
  double z = 1+*lambda, a = *lambda, r;
  int i=2;
  for(;i<1000;i++) {
    double e = *lambda/pow(i,*nu);
    a = a*e;
    z = z + a;
    if(e < 1) {
      /* relative truncation error */
      r = a/(1-e)/z;
      if(r < tol) break;
    }
  }
  *lambda = z;
  *nu = i;
}

/* A less accurate way to compute the mean. */
void compoisson_mean(double *lambda, double *nu)
{
  double tol = 1e-8;
  double z = 1+*lambda, a1 = (*lambda)*(*lambda)/pow(2,*nu), a2, r;
  double zx = *lambda, ax1 = 2*a1, ax2;
  double m,upper,lower;
  int i=3;
  if(*lambda == 0) {
    *nu = 0;
    return;
  }
  for(;i<1000;i++) {
    double e = *lambda/pow(i,*nu);
    double ex = *lambda/pow(i,*nu -1)/(i-1);
    a2 = a1*e;
    ax2 = ax1*ex;

    m = zx/z;
    upper = (zx + ax1/(1-ax2/ax1))/z;
    lower = zx/(z + a1/(1-a2/a1));

    if((ax2 < ax1) && (a2 < a1)) {
      /* relative truncation error */
      r = (upper-lower)/m;
      if(r < tol) break;
    }

    z = z + a1;
    zx = zx + ax1;
    a1 = a2;
    ax1 = ax2;
  }
  *lambda = m;
  *nu = i;
}

/* A less accurate way to compute the second moment. */
void compoisson_m2(double *lambda, double *nu)
{
  double tol = 1e-8;
  double z = 1+*lambda, a1 = (*lambda)*(*lambda)/pow(2,*nu), a2, r;
  double zxx = *lambda, axx1 = 4*a1, axx2;
  double m,upper,lower;
  int i=3;
  if(*lambda == 0) {
    *nu = 0;
    return;
  }
  for(;i<1000;i++) {
    double e = *lambda/pow(i,*nu);
    double exx = *lambda/pow(i,*nu -2)/(i-1)/(i-1);
    a2 = a1*e;
    axx2 = axx1*exx;

    m = zxx/z;
    upper = (zxx + axx1/(1-axx2/axx1))/z;
    lower = zxx/(z + a1/(1-a2/a1));

    if((axx2 < axx1) && (a2 < a1)) {
      /* relative truncation error */
      r = (upper-lower)/m;
      if(r < tol) break;
    }

    z = z + a1;
    zxx = zxx + axx1;
    a1 = a2;
    axx1 = axx2;
  }
  *lambda = m;
  *nu = i;
}
